﻿-- Author      : RisM
-- Create Date : 6/28/2009 4:04:27 PM

local SpeakinSpell = LibStub("AceAddon-3.0"):GetAddon("SpeakinSpell")
local L	= LibStub("AceLocale-3.0"):GetLocale("SpeakinSpell", false)

SpeakinSpell:PrintLoading("speech.lua")

-------------------------------------------------------------------------------
-- RANDOM SPEECHES - OnSpeechEvent, Speak or execute a command
-------------------------------------------------------------------------------


function SpeakinSpell:OnSpeechEvent( DetectedEventStub )
	local funcname = "OnSpeechEvent"

	-- fill in the stub
	local DetectedEvent = self:CreateDetectedEvent( DetectedEventStub )
	
	local SpeakForRankless = true
	
	--NOTE: it looks like I want to make a subroutine here and call it twice
	--		but the logic is actually slightly different between the ranked and rankless cases
	
	-- look for an event table entry for the ranked version
	if DetectedEvent.rankedkey then
		local de = self:CopyTable( DetectedEvent )
		de.key = de.rankedkey
		de.EventTableEntry = SpeakinSpellSavedData.EventTable[ de.key ]
		
		if SpeakinSpellSavedData.ShowSpellEvents or SpeakinSpell.DEBUG_MODE then
			self:Print( self:FormatSubs( "<EventTypePrefix><SpellLink>", de) )
		end

		if de.EventTableEntry then
			SpeakForRankless = false
			if self:AllowSpeakForSpell( de ) then
				self:SpeakForSpell( de )
			end
		else
			-- record a new event for this ranked key
			--self:DebugMsg(funcname, "no EventTableEntry for key:"..tostring(de.key))
			self:RecordNewEvent( de ) -- this updates the Create New window as necessary			
		end
	end
	
	-- check for a rankless version, but don't speak for both
	if DetectedEvent.ranklesskey then
		local de = self:CopyTable( DetectedEvent )
		de.key = de.ranklesskey
		de.EventTableEntry = SpeakinSpellSavedData.EventTable[ de.key ]

		if SpeakinSpellSavedData.EventTable[ de.key ] then
			if SpeakForRankless then
				if self:AllowSpeakForSpell( de ) then
					self:SpeakForSpell( de )
				end
			end
		else
			-- record a new RANKLESS event for this rankless key
			de.rank = "" -- to make the new event rankless -- NOTE: don't erase this before speaking, to maintain validity of <spellrank> variable
			--self:DebugMsg(funcname, "no EventTableEntry for key:"..tostring(de.key))
			self:RecordNewEvent( de ) -- this updates the Create New window as necessary
		end
	end

end



function SpeakinSpell:OnRecursiveEventDetected( text, DetectedEventStub, ParentDetectedEvent )
	local funcname = "OnRecursiveEventDetected"

	-- fill in the stub
	local de = self:CreateDetectedEvent( DetectedEventStub )
	
	-- NOTE: we should be able to use the default key for all recursive events
	--		shouldn't need to check for ranked vs. rankless keys
	
	-- set a parent link
	de.EventTableEntry = SpeakinSpellSavedData.EventTable[ de.key ]
	
	if SpeakinSpellSavedData.ShowSpellEvents or SpeakinSpell.DEBUG_MODE then
		self:Print( self:FormatSubs( "<EventTypePrefix><SpellLink>", de) )
	end

	if not de.EventTableEntry then
		self:ErrorMsg(funcname, "no event settings for "..tostring(text))
		return
	end

	if not self:AllowSpeakForSpell( de ) then
		return
	end
	
	local pop = true
	if not self.RuntimeData.ParentDetectedEvent then
		self.RuntimeData.ParentDetectedEvent = ParentDetectedEvent
	else
		-- add the new cascading event info the old one, in case it defines anything new
		self:ValidateObject( self.RuntimeData.ParentDetectedEvent, ParentDetectedEvent )
		pop = false
	end

	self:SpeakForSpell( de )

	if pop then
		self.RuntimeData.ParentDetectedEvent = nil	
	end
end


function SpeakinSpell:GetChatChannelForSpell(EventTableEntry)
	local funcname = "GetChatChannelForSpell"
	
	local key = self:GetScenarioKey()
	local channel = EventTableEntry.Channels[key]
	
	-- if the player wants to speak in the RAID_WARNING channel, 
	-- but doesn't have permission (not the leader or assist)
	-- then fall back to the RAID channel instead, so we say something at all
	-- instead of triggering an error message in the chat
	if (channel == "RAID_WARNING") and 
		not (IsRaidOfficer() or IsRaidLeader() or IsPartyLeader()) then
		self:DebugMsg(funcname, "You don't have permission to say Raid Warnings right now - using RAID chat instead")
		channel = "RAID"
	end
	
	return channel
end


-- returns a language that is usable in SendChatMessage
function SpeakinSpell:GetChatLanguageForSpell(EventTableEntry)
	local funcname = "GetChatLanguageForSpell"
	
	-- [X] Always Use Common
	-- or Common is the only known language
	local NumLanguages = GetNumLanguages() -- doesn't work during first-time load, so we need to be safe about it
	if not NumLanguages then
		NumLanguages = 1
	end
	if SpeakinSpellSavedData.AlwaysUseCommon or (NumLanguages < 2) then
		--self:DebugMsg(funcname, "always use common")
		return GetDefaultLanguage("player")
	end
	
	-- Player Selected "Random"
	if EventTableEntry.RPLanguage == L["Random"] then
		local Random = math.random(1,100)
		local Chance = EventTableEntry.RPLanguageRandomChance*100

		--self:DebugMsg(funcname, string.format(L["Random:%d <= Chance:%d"], Random, Chance) )
		if Random <= Chance then -- use racial language
			return self:GetRacialLanguage()
		else -- use Common
			return GetDefaultLanguage("player")
		end
	end
	
	--NOTE: GetDefaultLanguage("player") doesn't work during loader, or OnInitialize, or OnVariablesLoaded
	-- so we can't validate EventTableEntry.RPLanguage on-load
	--	(though it is safe for this function to return nil, with the same result because of SendChatMessage behavior)
	if not EventTableEntry.RPLanguage then
		EventTableEntry.RPLanguage = GetDefaultLanguage("player")
	end
	
	-- TODOFUTURE: add SpeakinSpell language filters like "Pirate" and "Drunk" (and treat Random as one of those?)
	--		if we're speakin in pirate, then this function should return GetDefaultLanguage("player")

	-- In all other cases, use the selected language as-is
	--self:DebugMsg(funcname, "selected:"..tostring(EventTableEntry.RPLanguage) )
	return EventTableEntry.RPLanguage
end


-- return true if this is a slash command
-- return false if this is other default text
function SpeakinSpell:ProcessChatSlashCommand(text, DetectedEvent)
	local funcname = "ProcessChatSlashCommand"
	
	if ( strsub(text, 1, 1) ~= "/" ) then
		return false
	end
	
	-- issue a warning if this a secure command like /cast
	-- If the string is in the format "/cmd blah", command will be "/cmd"
	local command = strmatch(text, "^(/[^%s]+)") or "";
	if IsSecureCmd(command) then
		self:Print(string.format(L["WARNING: can't execute protected command: %s"],text))
		return true
	end
	
	-- if this is a recursive call into a slash command, it needs some special handling
	-- do it with internal calls instead of passing through the global chat parser and back again
	if self:ProcessRecursiveUserMacro(text, DetectedEvent) then
		return true
	end
	
	self:DebugMsg(funcname,"executing slash command: " .. tostring(text) )
	
	-- create a hidden edit box to parse and send the slash command
	if not SpeakinSpellTempEditBox then
		SpeakinSpellTempEditBox = CreateFrame("EditBox", "SpeakinSpellTempEditBox", UIParent)
		SpeakinSpellTempEditBox:Hide()
	end
	
	-- send the command
	SpeakinSpellTempEditBox:SetText(text) 
	ChatEdit_SendText(SpeakinSpellTempEditBox)

	return true
end



function SpeakinSpell:SelfRaidWarn( text )
	-- show the line as a raid warning style popup that only I can see
	-- using the RaidWarningFrame makes this use the same font size and position
	-- the same as a real raid warning, except for the color, which is user-defined saved data
	local Color = SpeakinSpellSavedData.Colors.Channels["SELF RAID WARNING CHANNEL"]
	PlaySound("RaidWarning") --RaidNotice_AddMessage doesn't play sound
	RaidNotice_AddMessage(RaidWarningFrame, text, Color)
end


function SpeakinSpell:SelfChat_SpeakinSpellChannel( text )
	if not SpeakinSpell.Colors.Channels then
		self:LoadChatColorCodes()
	end
	local msg = tostring(SpeakinSpell.Colors.Channels["SPEAKINSPELL CHANNEL"]).."SpeakinSpell|r: " .. tostring(text)
	DEFAULT_CHAT_FRAME:AddMessage( msg )
end


-- SelfChat_BossWhisper
-- mimic boss whispers
-- text = "[voice] whispers: something"
-- this function does not prefix anything
-- TODO: distribute to all appropriate chat frames per player's filter settings and custom chat frames
function SpeakinSpell:SelfChat_BossWhisper( text )
	local Color = ChatTypeInfo["RAID_BOSS_WHISPER"]
	DEFAULT_CHAT_FRAME:AddMessage( text, Color.r, Color.g, Color.b, Color.id )
end


-- "[Mysterious Voice] whispers: " .. text
-- TODO: distribute to all MONSTER_WHISPER chat frames per player's filter settings and custom chat frames
function SpeakinSpell:SelfChat_MysteriousVoice( text )
	--local MONSTER_WHISPER = ChatTypeInfo["MONSTER_WHISPER"]
	if not SpeakinSpell.Colors.Channels then
		self:LoadChatColorCodes()
	end
	local Color = SpeakinSpellSavedData.Colors.Channels["MYSTERIOUS VOICE"]
	local msg = L["[Mysterious Voice] whispers: "] .. tostring(text)
	DEFAULT_CHAT_FRAME:AddMessage( msg, Color.r, Color.g, Color.b --[[, MONSTER_WHISPER.id--]] )
end


function SpeakinSpell:EmoteToSay( line )
	-- line is "emotes in the chat"
	-- return value wants to be "* <player> emotes in the chat *"
	return string.format( L["* %s %s *"], UnitName("player"), line ) --TODOFUTURE: make this string format user-configurable
end


-- does not format subs
-- SendChatMessage("msg" 
--		[,"chatType=say,party,whisper,etc" 
--		[,"language=nil,common,dwarvish,etc" 
--		[,"channel means whisper target name or channel number 1,2,3"]]]);
function SpeakinSpell:SayOneLine(line, channel, Language, EnableWhisperTarget, DetectedEvent)
	-- never print empty lines
	if (not line) or (line == "") then
		return
	end
	
	-- if this message is a slash command, send it through an editBox function to parse and execute the command
	--NOTE: a specified slash command overrides the default channel and whisper features below
	--		it doesn't make sense to combine them
	if self:ProcessChatSlashCommand(line, DetectedEvent) then
		return
	end
	
	-- define custom channel functions
	local ChannelFuncs = {
		["SPEAKINSPELL CHANNEL"]	= function(line)	self:SelfChat_SpeakinSpellChannel(line)	end,
		["RAID_BOSS_WHISPER"]		= function(line)	self:SelfChat_BossWhisper(line)			end,
		["MYSTERIOUS VOICE"]		= function(line)	self:SelfChat_MysteriousVoice(line)		end,
		["SELF RAID WARNING CHANNEL"]= function(line)	self:SelfRaidWarn(line)					end,
	}
		
	-- say the message to the chat, if applicable
	if channel then
		-- check for proprietary SS-created channels
		local Func = ChannelFuncs[ channel ]
		if Func then
			Func( line )
		else
			-- all others go directly through Blizzard's SendChatMessage
			--TODOFUTURE: should this check for "|" for invalid escape sequences to prevent LUA errors? or allow color changes and stuff?
			SendChatMessage(line, channel, Language);
		end
	end
	
	-- send the message to a whisper to the target, if applicable
	if EnableWhisperTarget then
		-- NOTE: always use Common(nil) for the language just to ensure the person you whisper can read what you say
		-- NOTE: for whispered emotes, reformat the whisper to sound like the way /e works
		if channel == "EMOTE" then
			line = self:EmoteToSay(line)
		end
		SendChatMessage(line, "WHISPER", nil, DetectedEvent.target);
	end
end


-- expand recursive /ss macro calls
-- so that we have an even random distribution across all shared sub-lists
function SpeakinSpell:ExpandSSMacros( ExpandedMessages, EventTableEntry )
	for _,msg in pairs( EventTableEntry.Messages ) do
		--NOTE: GetETEForSSMacro only works if the msg is a single one-line macro command that is defined
		--		for multi-line "/ss macro A<newline>/ss macro B" GetETEForSSMacro will return nil
		--		and we'll leave it as-is
		--		likewise for any speech that is not a /ss macro command
		local ete = self:GetETEForSSMacro(msg)
		if ete then
			self:ExpandSSMacros( ExpandedMessages, ete )
		else
			tinsert( ExpandedMessages, msg )
		end
	end
end


-- replace a macro call with the macro's speech list
-- NOTE: expand it out in-place instead of appending to the end
--		the recursive ExpandSSMacros function above is complete/automatic
--		this function is meant for manual editing in the GUI
function SpeakinSpell:ExpandOneSSMacro( Messages, index )
	-- find the event table entry for this message
	local msg = Messages[ index ]
	if not msg then
		self:ErrorMsg(funcname, "msg is nil")
		return
	end
	local ete = SpeakinSpell:GetETEForSSMacro(msg)
	if not ete then
		self:ErrorMsg(funcname, "ete is nil")
		return
	end

	-- capture everything up to the index, but not the index
	NewMessages = {}
	for i = 1, index-1 do
		tinsert( NewMessages, Messages[i] ) --adds to the end of NewMessages at highest unused index
	end
	-- NewMessages[ index ] = nil
	
	-- add the /ss macro command speeches
	for _,msg in pairs( ete.Messages ) do
		tinsert( NewMessages, msg )
	end
	
	-- add the rest of the original array
	for i = index+1, #Messages do
		tinsert( NewMessages, Messages[i] ) --adds to the end of NewMessages at highest unused index
	end
	
	-- remove duplicates while we go out
	return self:StringArray_Compress( NewMessages )
end


function SpeakinSpell:GetRandomMessage( DetectedEvent )
	--TODO: find a better solution than this - see below - thinking "/ss group mount"
	return self:GetRandomTableEntry( DetectedEvent.EventTableEntry.Messages, self:GetLastMessage( DetectedEvent ) )
	--[[
	-- expand recursive /ss macro calls
	-- so that we have an even random distribution across all shared sub-lists
	local ExpandedMessages = {}
	self:ExpandSSMacros( ExpandedMessages, DetectedEvent.EventTableEntry )
	-- if embedded macro calls include repeats, filter those out to avoid giving them extra weight in the random selection
	ExpandedMessages = self:StringArray_Compress( ExpandedMessages )
	return self:GetRandomTableEntry( ExpandedMessages, self:GetLastMessage( DetectedEvent ) )
	--]]
end


function SpeakinSpell:SpeakForSpell(DetectedEvent)
	local funcname = "SpeakForSpell"
	
	-- determine which channel we will speak in
	local channel = self:GetChatChannelForSpell(DetectedEvent.EventTableEntry)
	
	--TODOFUTURE: make it an option to whisper yourself - but be wary of the auto self-cast logic
	local EnableWhisperTarget = DetectedEvent.EventTableEntry.WhisperTarget and DetectedEvent.target and (DetectedEvent.target ~= "") and (DetectedEvent.target ~= "Unknown") and (DetectedEvent.target ~= UnitName("player")) and UnitIsFriend("player",DetectedEvent.target)

	if not channel and not EnableWhisperTarget then
		-- this spell's speeches have no target for the current situation
		-- NOTE: this rule applies even if a slash command override is present in the randomly selected message text
		self:DebugMsg(funcname, self:FormatSubs("<SpellLink> is silent while <Scenario>",DetectedEvent) )
		return
	end
	
	-- select a random speech
	local msg = self:GetRandomMessage( DetectedEvent )
	if not msg then
		self:DebugMsg(funcname, self:FormatSubs("<SpellLink> doesn't have any speeches defined",DetectedEvent) )
		return
	end
	
	-- record the time when we spoke for this spell, along with the last target, and last message spoken
	-- to support duplicate prevent, announcement limits, and cooldowns
	-- NOTE: record the unprocessed message text, so that changing the value of substitution variables
	--		does not make us think it's a new random message selection
	self:SaveAnnouncementHistory( DetectedEvent, msg )
	
	-- format substitutions (including <newline>)
	msg = self:FormatSubs( msg, DetectedEvent )
	
	-- break up the message into separate lines, if applicable
	local lines = { strsplit("\n",msg) }

	-- determine the language: Common vs. Dwarvish, etc.
	local Language = self:GetChatLanguageForSpell( DetectedEvent.EventTableEntry )

	for _,line in ipairs(lines) do
		-- process substitution variables
		self:SayOneLine(line, channel, Language, EnableWhisperTarget, DetectedEvent)
	end
end

